//----------------------------------------------------------------------------//
// RbPSe - Sony Play Station Emulator                                         //
// Copyright (C) notes:                                                       //
//   Code: RomikB, 2002-2003                                                  //
//        <romikb@mail.ru>                                                    //
//                                                                            //
// This program is free software; you can redistribute it and/or              //
// modify it under the terms of the GNU General Public License                //
// as published by the Free Software Foundation; either version 2             //
// of the License, or (at your option) any later version.                     //
//                                                                            //
// This program is distributed in the hope that it will be useful,            //
// but WITHOUT ANY WARRANTY; without even the implied warranty of             //
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              //
// GNU General Public License for more details.                               //
//                                                                            //
// You should have received a copy of the GNU General Public License          //
// along with this program; if not, write to the Free Software                //
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.//
//----------------------------------------------------------------------------//
/******************************************************************************/
#include "include/rbpse.sdk.h"
#include "interface.h"
/******************************************************************************/
// Include
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
//
#if   defined(__SDL_SUPPORT__)
# include <SDL/SDL.h>
#endif
/******************************************************************************/
typedef struct {
	long	library_handle;
	int		(*PluginInit)(RbPSeDataStruc*,RbPSeHookStruc*,RbPSeFuncStruc*,RbPSeStartStruc*);
	int		(*PluginOpen)(RbPSeStartStruc*);	//need to change
	int		(*PluginStart)();
	int		(*PluginStop)();
	int		(*PluginSave)();
	int		(*PluginLoad)();
	int		(*PluginClose)();
	int		(*PluginShutdown)();
} RbPSePluginFunctions;
/******************************************************************************/
// Data
const char *IniFile = "RbPSe.cfg";
const char *library_section = "library";
const char *plugins_section = "plugins";
//const char *interface_function_names[]={"EmulatorInterface",0};
const char *plugin_function_names[] =
{
	"RbPSePluginInit",
	"RbPSePluginOpen",
	"RbPSePluginStart",
	"RbPSePluginStop",
	"RbPSePluginSave",
	"RbPSePluginLoad",
	"RbPSePluginClose",
	"RbPSePluginShutdown",
	0
};
const char *library_names[] =
{
	"emulog",
	"memory",
	"hardware",
	"cpudyn",
	0
};
const char *plugins_names[] =
{
	"00","01","02","03","04","05","06","07","08","09","10","11","12","13","14",
	"15","16","17","18","19","20","21","22","23","24","25","26","27","28",
	0
};
char		library[4][32];
char		plugins[32][32];
RbPSePluginFunctions	library_functions[4];
RbPSePluginFunctions	plugins_functions[32];
//long		interface_functions[1][2];
long		library_count = 4;
long		plugins_count = 0;
long		emu_thread = 0;
long		cur_core_state = CORE_STATE_NONE;
//long		is_core_wait_for_stop;
message_function	key_handlers_[32];
long		key_handlers_count_=0;
/******************************************************************************/
int		CoreInit(void);
int		CoreOpen(void);
int		CoreStart(void);
int		CoreStop(void);
int		CoreSave(void);
int		CoreLoad(void);
int		CoreClose(void);
int		CoreShutdown(void);
/******************************************************************************/
int print_help(){
	char ch;
	printf("RbPSe - Sony Play Station Emulator (www.rbpse.sf.net)\n\
Version 0.40 Usage:\n\tRbPSe Option [Advanced Options]\nOptions:\n\
\t--cdrom device\t\t\tExecute CD-ROM device\n\
\t--exe filename\t\t\tExecute PSX-EXE file\n\
\t--iso filename\t\t\tExecute CD-ROM ISO image\n\
\t--about\t\t\t\tPrint About Information\nAdvanced Options:\n\
\t--current dir\t\t\tUse \"dir\" as Current directory\n\
\t--profile dir\t\t\tUse \"dir\" as Profile directory\n\
\t--plugins dir\t\t\tUse \"dir\" as Plugins directory\n\
\t--logfile dir\t\t\tUse \"dir\" as Logfile directory\n\
\t--interface filename\t\tUse interface library \"filename\"\n");
	printf("\nPress Enter...");
	scanf("%c",&ch);
	return -1;
}
/******************************************************************************/
int print_about(){
	printf("RbPSe - Sony Play Station Emulator (www.rbpse.sf.net)\n\
    -    Version 0.40(.0) 2003/05/01\n\
    -    Developers:\n\
             RomikB aka Roman Zakharchuk\n\
    -    Testers:\n\
             Organic\n\
             Lynx\n\
    -    Thanks to:\n\
             Organic,Lynx,PSEmu Plugin Developers,Jeksik\n");
	return 0;
}
/******************************************************************************/
long CoreFunction(void *data){
    __asm__("								\n\t\
    			movl %%ecx,%%ebx			\n\t\
                leal (%%ebx,%%edx,4),%%ecx	\n\t\
                jmp *%%ecx					\n\t\
    		"
            :
            : "S"(RbPSeDataData.Data___Table___),
              "D"(RbPSeDataData.Data___Memory___),
              "c"(RbPSeDataData.Data___Dynarec___),
              "d"(RbPSeStartData.ExecAddr)
    );
	return 0;
}
/******************************************************************************/
int CoreInit(void){
	int i,j;
	long lvar,count;
	lvar=RbPSe_fopen(RbPSeStartData.ProfileDir,IniFile,0);
	if(!lvar){
		printf("[!!!!!!!]Error: Cann't open config file \"%s\"\n",IniFile);
		return -1;
	}
	RbPSe_fclose(lvar);
	count=RbPSe_ReadStrSettings(RbPSeStartData.ProfileDir,IniFile,
		library_section,library_names,"_nf_",library);
    if(count!=4) return -1;
	for(i=0;i<4;i++){
		if(!strcmp(library[i],"_nf_")){
			printf("[!!!!!!!]Error: Library not defined \"%s\"\n",
					library_names[i]);
			return -1;
		}
		lvar=RbPSe_LoadLibrary(RbPSeStartData.PluginsDir,
				library[i],plugin_function_names,(long*)&library_functions[i]);
		if(lvar){
			printf("[!!!!!!!]Error: Cann't load library \"%s\"\n",library[i]);
			return -1;
		}
		if(library_functions[i].PluginInit(&RbPSeDataData,&RbPSeHookData,&RbPSeFuncData,&RbPSeStartData))
        {
            printf("[!!!!!!!]Error: Cann't init library \"%s\"\n",library[i]);
            return -1;
        }
	}
	count=RbPSe_ReadStrSettings(RbPSeStartData.ProfileDir,IniFile,
		plugins_section,plugins_names,"_nf_",plugins);
	for(i=0,j=0;i<count;i++){
		if(!strcmp(plugins[i],"_nf_")) continue;
		lvar=RbPSe_LoadLibrary(RbPSeStartData.PluginsDir,
				plugins[i],plugin_function_names,(long*)&plugins_functions[i]);
		if(lvar){
			printf("[!!!!!!!]Warning: Cann't load plugin \"%s\"\n",plugins[i]);
			continue;
		}
		plugins_functions[j].PluginInit(&RbPSeDataData,&RbPSeHookData,&RbPSeFuncData,&RbPSeStartData);
		j++;
	}
	plugins_count=j;
	return 0;
}
/******************************************************************************/
int CoreOpen(void){
	int i,KbHndr;
	for(i=0;i<4;i++){
		library_functions[i].PluginOpen(&RbPSeStartData);
	}
    key_handlers_count_ = 0;
	for(i=0;i<plugins_count;i++){
		KbHndr = plugins_functions[i].PluginOpen(&RbPSeStartData);
        if(KbHndr == -1){
        	printf("[!!!!!!!]Warning: Cann't open plugin...\n");
        }else{
        	if(KbHndr){
            	key_handlers_[key_handlers_count_++] = (message_function)KbHndr;
            }
        }
	}
    if(!RbPSeStartData.ExecAddr){
    	printf("[!!!!!!!]Error: Cann't find plugin for this source type\n");
        return -1;
    }
	emu_thread=RbPSe_CreateThread(CoreFunction,0);
	return 0;
}
/******************************************************************************/
int CoreStart(void){
	int i;
	for(i=0;i<4;i++){
		library_functions[i].PluginStart();
	}
	for(i=0;i<plugins_count;i++){
		plugins_functions[i].PluginStart();
	}
	RbPSe_ResumeThread(emu_thread);
	return 0;
}
/******************************************************************************/
int CoreStop(void){
	int i;
	RbPSe_SuspendThread(emu_thread);
	for(i=0;i<4;i++){
		library_functions[i].PluginStop();
	}
	for(i=0;i<plugins_count;i++){
		plugins_functions[i].PluginStop();
	}
	return 0;
}
/******************************************************************************/
int CoreSave(void){
	return 0;
}
/******************************************************************************/
int CoreLoad(void){
	return 0;
}
/******************************************************************************/
int CoreClose(void){
	int i;
	RbPSe_KillThread(emu_thread);
	for(i=0;i<4;i++){
		library_functions[i].PluginClose();
	}
	for(i=0;i<plugins_count;i++){
		plugins_functions[i].PluginClose();
	}
	return 0;
}
/******************************************************************************/
int CoreShutdown(void){
	int i;
	for(i=0;i<4;i++){
		library_functions[i].PluginShutdown();
	}
	for(i=0;i<plugins_count;i++){
		plugins_functions[i].PluginShutdown();
	}
	return 0;
}
/******************************************************************************/
int main(int argc,char **argv)
{
	int i=3;
	char *cur_dir=0;
	if(argc==1) return print_help();
	while(1)
    {
		if(!strcmp(argv[1],"--cdrom"))
        {
			if(2<argc)
            {
				RbPSeStartData.SourceType=0;
				RbPSeStartData.SourceName=malloc(256);
				strcpy(RbPSeStartData.SourceName,argv[2]);
				break;
			}
            else
            {
				return print_help();
			}
		}
		if(!strcmp(argv[1],"--iso"))
        {
			if(2<argc)
            {
				RbPSeStartData.SourceType=1;
				RbPSeStartData.SourceName=malloc(256);
				strcpy(RbPSeStartData.SourceName,argv[2]);
				break;
			}
            else
            {
				return print_help();
			}
		}
		if(!strcmp(argv[1],"--exe"))
        {
			if(2<argc)
            {
				RbPSeStartData.SourceType=2;
				RbPSeStartData.SourceName=malloc(256);
				strcpy(RbPSeStartData.SourceName,argv[2]);
				break;
			}
            else
            {
				return print_help();
			}
		}
        if(!strcmp(argv[1],"--about")) return print_about();
		return print_help();
	}
	RbPSeStartData.PluginsDir=malloc(256);
	RbPSeStartData.ProfileDir=malloc(256);
	RbPSeStartData.LogfileDir=malloc(256);
	*(RbPSeStartData.PluginsDir)=0;
	*(RbPSeStartData.ProfileDir)=0;
	*(RbPSeStartData.LogfileDir)=0;
	while(1)
    {
		if(i<argc)
        {
			if(!strcmp(argv[i],"--current"))
            {
				i++;
				if(i<argc)
                {
					cur_dir=malloc(256);
					strcpy(cur_dir,argv[i]);
					i++;
					continue;
				}
                else
                {
					return print_help();
				}
			}
			if(!strcmp(argv[i],"--plugins"))
            {
				i++;
				if(i<argc)
                {
					strcpy(RbPSeStartData.PluginsDir,argv[i]);
					i++;
					continue;
				}
                else
                {
					return print_help();
				}
			}
			if(!strcmp(argv[i],"--profile"))
            {
				i++;
				if(i<argc)
                {
					strcpy(RbPSeStartData.ProfileDir,argv[i]);
					i++;
					continue;
				}
                else
                {
					return print_help();
				}
			}
			if(!strcmp(argv[i],"--logfile"))
            {
				i++;
				if(i<argc)
                {
					strcpy(RbPSeStartData.LogfileDir,argv[i]);
					i++;
					continue;
				}
                else
                {
					return print_help();
				}
			}
            return print_help();
		}
        else
        {
			break;
		}
	}
	if(cur_dir) chdir(cur_dir);
	if(!(*(RbPSeStartData.PluginsDir)))
    {
		if(!chdir("Plugins"))
        {
        	getcwd(RbPSeStartData.PluginsDir,256);
            chdir("..");
        }
        else
        {
        	printf("[???????]Default Plugins directory not found. Use \"--plugins\".\n");
            return -1;
        }
	}
	if(!(*(RbPSeStartData.ProfileDir)))
    {
    	if(!chdir("Profiles"))
        {
        	if(!chdir("default"))
            {
            	getcwd(RbPSeStartData.ProfileDir,256);
                chdir("..");
            }
            else
            {
				printf("[???????]Default Profile directory not found. Use \"--profile\".\n");
				return -1;
            }
            chdir("..");
        }
        else
        {
            printf("[???????]Default Profile directory not found. Use \"--profile\".\n");
            return -1;
        }
	}
	if(!(*(RbPSeStartData.LogfileDir)))
    {
		getcwd(RbPSeStartData.LogfileDir,256);
	}
	switch(RbPSeStartData.SourceType)
    {
	case 0:
		printf("CD-ROM\n");
		break;
	case 1:
		printf("ISO image\t: %s\n",RbPSeStartData.SourceName);
		break;
	case 2:
		printf("PSX-EXE\t\t: %s\n",RbPSeStartData.SourceName);
		break;
	}
	printf("Plugins at\t: %s\nProfile at\t: %s\nLogfile at\t: %s\n",
		RbPSeStartData.PluginsDir,RbPSeStartData.ProfileDir,RbPSeStartData.LogfileDir);

	RbPSeStartData.ExecAddr = 0;

	InterfaceInit(&(RbPSeStartData.Display));

	if(CoreInit()){
		printf("[!!!!!!!]Error: CoreInit Unsuccessful\n");
		return InterfaceShutdown(-2);
	}
	if(CoreOpen()){
		printf("[!!!!!!!]Error: CoreOpen Unsuccessful\n");
		return InterfaceShutdown(-2);
	}
	if(CoreStart()){
		printf("[!!!!!!!]Error: CoreStart Unsuccessful\n");
		return InterfaceShutdown(-2);
	}

    while(1){
		switch(InterfaceMessage(key_handlers_,key_handlers_count_,cur_core_state)){
	    case INTERFACE_MESSAGE_EXIT:
			if(CoreClose()){
				printf("[!!!!!!!]Error: CoreClose Unsuccessful\n");
				return InterfaceShutdown(-2);
			}
			if(CoreShutdown()){
				printf("[!!!!!!!]Error: CoreShutdown Unsuccessful\n");
				return InterfaceShutdown(-2);
			}
			return InterfaceShutdown(0);
        case INTERFACE_MESSAGE_STOP:
            if(CoreStop()){
                printf("[!!!!!!!]Error: CoreStop Unsuccessful\n");
                return InterfaceShutdown(-2);
            }
            break;
	    default:
        	break;
        }
    }
	return InterfaceShutdown(0);
}

